// ANNA - Anna is Not Nothingness Anymore                                                         //
//                                                                                                //
// (c) Copyright 2005-2015 Eduardo Ramos Testillano & Francisco Ruiz Rayo                         //
//                                                                                                //
// See project site at http://redmine.teslayout.com/projects/anna-suite                           //
// See accompanying file LICENSE or copy at http://www.teslayout.com/projects/public/anna.LICENSE //


/*
  Ejemplo de programa servidor. Atiende peticiones aritmeticas, el protocolo de transporte sera HTTP
  ver http::Transport y el contenido del mensaje sera el resutaldo de un comm::Codec con los
  (x, y, op) -> El resultado sera estos tres componente mas result.

  Para poder probar el sistema de congestion se puede indicar un numero de milisegundos de
  retardo aplicados a cada contestacion.
  
  Los clientes pueden ser: wims20_kclient.p wims20_client.p
*/
#include <iostream>

#include <anna/core/core.hpp>

#include <anna/xml/Node.hpp>
#include <anna/xml/Attribute.hpp>

#include <anna/comm/comm.hpp>

#include <anna/http/Request.hpp>
#include <anna/http/Response.hpp>
#include <anna/http/Handler.hpp>
#include <anna/http/Transport.hpp>
#include <anna/http/functions.hpp>

#include <anna/http/wims20/ServerSide.hpp>

using namespace std;

class MyCommunicator : public comm::Communicator {
public:
   MyCommunicator () {;}
};

class MyHandler : public http::Handler {
public:
   MyHandler () : http::Handler ("wims20_rserver::MyHandler"), a_request (NULL) {;}

   static const char* className () throw () { return "wims20_rserver::ReceiverFactory"; }

private:
   MyCommunicator* a_communicator;
   http::wims20::ServerSide* a_request;

   void initialize () throw (RuntimeException);
   void evRequest (ClientSocket&, const http::Request& request) throw (RuntimeException);
   void evResponse (ClientSocket&, const http::Response&) throw (RuntimeException) {;}
};

class HTTPArithmeticServer : public comm::Application {
public:
   HTTPArithmeticServer ();
      
private:
   MyCommunicator a_communicator;
   ReceiverFactoryImpl <MyHandler> a_receiverFactory;
   comm::ServerSocket* a_serverSocket;

   void initialize () throw (RuntimeException);
   void run () throw (RuntimeException);
};

using namespace std;
using namespace anna::comm;

int main (int argc, const char** argv)
{
   CommandLine& commandLine (CommandLine::instantiate ());
   HTTPArithmeticServer app;

   http::functions::initialize ();

   srand (time (NULL));

   try {
      commandLine.initialize (argv, argc);
      commandLine.verify ();

      Logger::setLevel (Logger::Debug); 
      Logger::initialize ("wims20_server", new anna::TraceWriter ("file.trace", 4048000));
 
      app.start ();
   }
   catch (Exception& ex) {
      cout << ex.asString () << endl;
   }
   
   return 0;
}

HTTPArithmeticServer::HTTPArithmeticServer () : 
   Application ("wims20_rserver", "Servidor HTTP/WIMS 2.0 de operaciones aritmeticas (iRS)", "1.0") 
{
   CommandLine& commandLine (CommandLine::instantiate ());
      
   commandLine.add ("p", CommandLine::Argument::Mandatory, "Puerto en el que atender peticiones");
   commandLine.add ("a", CommandLine::Argument::Mandatory, "Direccion IP en la que atender");
   commandLine.add ("trace", CommandLine::Argument::Optional, "Nivel de trazas (debug,warning, error,...)");
   commandLine.add ("domain", CommandLine::Argument::Mandatory, "Domain indicado en la peticion WIMS 2.0");
   commandLine.add ("path", CommandLine::Argument::Optional, "Path indicado en la peticion WIMS 2.0");
}

void HTTPArithmeticServer::initialize () 
   throw (RuntimeException)
{
   CommandLine& cl (CommandLine::instantiate ());

   int port = cl.getIntegerValue ("p");
   const comm::Device* device = Network::instantiate ().find (Device::asAddress (cl.getValue ("a")));

   comm::TransportFactory* ttff = &http::Transport::getFactory ();

   a_serverSocket = new ServerSocket (INetAddress (device, port), cl.exists ("r"), ttff);
   a_serverSocket->setReceiverFactory (a_receiverFactory);
}

void HTTPArithmeticServer::run ()
   throw (RuntimeException)
{
   CommandLine& cl (CommandLine::instantiate ());

   a_communicator.attach (a_serverSocket);

   if (cl.exists ("trace"))
      Logger::setLevel (Logger::asLevel (cl.getValue ("trace")));

   a_communicator.accept ();
}

void MyHandler::initialize ()
   throw (RuntimeException)
{
   CommandLine& cl (CommandLine::instantiate ());

   string domain = cl.getValue ("domain");

   if (cl.exists ("path")) {
      string path = cl.getValue ("path");
      a_request = new http::wims20::ServerSide (domain, path);
   }
   else
      a_request = new http::wims20::ServerSide (domain);
  
   a_communicator = app::functions::component <MyCommunicator> (ANNA_FILE_LOCATION);
   allocateResponse ()->createHeader (http::Header::Type::Date); 
}

// Recibe una peticin de operacin aritmtica y visualiza el resultado por pantalla.
// Para mantener la simplicidad del ejemplo slo notifica al cliente si ha ido bien 
// o no, pero no deuuelve ningn resultado
//
// Definimos que nuestro servicio deber recibir una peticin REST de la forma:
//   http://<domain>/<path>/math/guest@tid.es/<op>?X=nnn&Y=nnn
// Observar que la <op> correspondera con las parte other_possible_levels
// de la especificacin WIMS 2.0
void MyHandler::evRequest (ClientSocket& clientSocket, const http::Request& request)
   throw (RuntimeException)
{
   LOGMETHOD (TraceMethod tm ("MyReceiver", "apply", ANNA_FILE_LOCATION));

   a_request->decode (request);

   cout << a_request->asString () << endl << endl;

   http::Response* response = allocateResponse ();

   if (a_request->getServiceID () != "math" || a_request->getGUID () != "user@tid.es") {
      response->setStatusCode (401);
      clientSocket.send (*response);
      return;
   }

   if (a_request->hasOtherLevels () == false) {
      response->setStatusCode (416);
      clientSocket.send (*response);
      return;
   }

   // El primer nivel indica la operacin a realizar
   const std::string& op = *http::wims20::ServerSide::otherLevel (a_request->other_level_begin ());

   int x, y;

   try {
      x = a_request->getIntegerValue ("X");
      y = a_request->getIntegerValue ("Y");
      response->setStatusCode (200);

      if (op == "sum") 
         cout << x << " + " << y << " = " << x + y << endl << endl;
      else if (op == "sub") 
         cout << x << " - " << y << " = " << x - y << endl << endl;
      else if (op == "mul") 
         cout << x << " * " << y << " = " << x * y << endl << endl;
      else if (op == "div") {
         if (y != 0)
            cout << x << " / " << y << " = " << x / y << endl << endl;
         else {
            response->setStatusCode (400);
            response->setReasonPhrase ("Division por cero");
         }
      }
      else 
         response->setStatusCode (501);
   }
   catch (RuntimeException& ex) {
      ex.trace ();
      response->setStatusCode (500);
      response->setReasonPhrase (ex.getText ());
   }
   
   try {
      clientSocket.send (*response);
   }
   catch (Exception& ex) {
      ex.trace ();
   }
}

