// ANNA - Anna is Not Nothingness Anymore                                                         //
//                                                                                                //
// (c) Copyright 2005-2015 Eduardo Ramos Testillano & Francisco Ruiz Rayo                         //
//                                                                                                //
// See project site at http://redmine.teslayout.com/projects/anna-suite                           //
// See accompanying file LICENSE or copy at http://www.teslayout.com/projects/public/anna.LICENSE //


#ifndef anna_core_util_CommaSeparatedList_hpp
#define anna_core_util_CommaSeparatedList_hpp

namespace anna {

/**
   Obtiene una lista de objetos a partir de la lista expresada con el operador ','.
   Lo mas normal es que sea usada para pasar un numero indeterminado de parametros del mismo tipo a un metodo
   o funcion.

   Por ejemplo el modo de uso seria el siguiente:

   \code
      CommaSeparatedList <Objeto> params;
      Objeto a, b, c;

      params = a, b, c;

      for (CommaSeparatedList<Objeto>::const_iterator ii = params.begin (), maxii = params.end (); ii != maxii; ii ++) {
         Objecto* obj = CommaSeparatedList<Objeto>::data (ii);

         .... usa el objeto de la lista ...
      }

   \endcode

   Un ejemplo mas realista. Suponed que tenemos una funcion que tiene que recibir un numero indeterminado
   de instancia de tipo Objeto. Se definira de la forma:

   \code
   void cualquierFuncion (const anna::CommaSeparatedList <Objeto>& csl) {
      for (CommaSeparatedList<Objeto>::const_iterator ii = params.begin (), maxii = params.end (); ii != maxii; ii ++) {
         Objecto* obj = CommaSeparatedList<Objeto>::data (ii);

         .... usa el objeto de la lista ...
      }
   }
   \endcode

   La invocacion a la funcion sera:

   \code
   Objeto aaa, bbb, ccc;

   cualquierFunction (aaa, bbb, ccc);
   cualquierFunction (bbb, ccc);

   \endcode

   En el primer caso la lista de objetos terminara conteniendo los elementos aaa, bbb y ccc..
   En el segundo caso la lista de objetos terminara conteniendo los elementos bbb y ccc.
*/
template <typename T> class CommaSeparatedList {
public:
  typedef typename std::vector <T*>::iterator iterator; /**< Definicion para recorrer los elementos. */
  typedef typename std::vector <T*>::const_iterator const_iterator; /**< Definicion para recorrer los elementos */

  /**
     Constructor vacio.
  */
  CommaSeparatedList() {;}

  /**
     Destructor.
  */
  ~CommaSeparatedList() { a_parameters.clear(); }

  /**
     Operador coma re-escrito para recorrer los objetos de la expresion basada en ','.
     Cada uno de los elementos de la expresion es convertido a un elemento de la lista.
     \param t Parametro con el que ampliar la lista.
     \return La instancia de la lista destino.
  */
  CommaSeparatedList<T>& operator, (T& t) throw() { a_parameters.push_back(&t); return *this; }

  /**
     Operador de asignacion.
     Amplia la lista con el elemento recibido.
     \param t Parametro con el que ampliar la lista.
     \return La instancia de la lista destino.
  */
  CommaSeparatedList<T>& operator= (T& t) throw() { a_parameters.push_back(&t); return *this; }

  /**
     Operador de copia.
     \param rsh Lista de parametros a copiar.
     \return La instancia de la lista destino.
  */
  CommaSeparatedList<T>& operator= (CommaSeparatedList<T>& rsh) throw() {
    if(this != &rsh)
      a_parameters = rsh.parameters;

    return *this;
  }

  /**
     Elimina el contenido actual de la lista de parametros.
  */
  void clear() throw() { a_parameters.clear(); }

  /**
     Devuelve un iterator al primer parametro de la lista.
     \return Un iterator al primer parametro de la lista.
  */
  iterator begin() throw() { return a_parameters.begin(); }

  /**
     Devuelve un iterator al primer parametro de la lista.
     \return Un iterator al primer parametro de la lista.
  */
  const_iterator begin() const throw() { return a_parameters.begin(); }

  /**
     Devuelve un iterator al ultimo parametro de la lista.
     \return Un iterator al ultimo parametro de la lista.
  */
  iterator end() throw() { return a_parameters.end(); }

  /**
     Devuelve un iterator al ultimo parametro de la lista.
     \return Un iterator al ultimo parametro de la lista.
  */
  const_iterator end() const throw() { return a_parameters.end(); }

  /**
     Devuelve el puntero sobre el que esta posicionado el iterador recibido como parametro.
     \return El puntero sobre el que esta posicionado el iterador recibido como parametro.
  */
  static T* data(iterator& ii) throw() { return *ii; }

  /**
     Devuelve el puntero sobre el que esta posicionado el iterador recibido como parametro.
     \return El puntero sobre el que esta posicionado el iterador recibido como parametro.
  */
  static const T* data(const_iterator& ii) throw() { return *ii; }

private:
  std::vector <T*> a_parameters;
};

}

#endif

