// ANNA - Anna is Not Nothingness Anymore                                                         //
//                                                                                                //
// (c) Copyright 2005-2015 Eduardo Ramos Testillano & Francisco Ruiz Rayo                         //
//                                                                                                //
// See project site at http://redmine.teslayout.com/projects/anna-suite                           //
// See accompanying file LICENSE or copy at http://www.teslayout.com/projects/public/anna.LICENSE //


#ifndef anna_diameter_comm_Response_hpp
#define anna_diameter_comm_Response_hpp


#include <anna/core/util/Recycler.hpp>
#include <anna/core/Allocator.hpp>
#include <anna/core/RuntimeException.hpp>

// Local
#include <anna/diameter/defines.hpp>
#include <anna/diameter.comm/ClassCode.hpp>


namespace anna {
class DataBlock;
}


namespace anna {

namespace diameter {

namespace comm {

class Session;
class Timer;
class Message;

/**
   Answers that we could receive from diameter servers.

   La respuesta correspondiente a una peticion se genera automaticamente al invocar a diameter::comm::Session::send.
   ANNA::diameter::comm notifies el resultado de la operacion solicitada mediante la invocacion
   al metodo-menejador diameter::comm::Session::eventResponse de nuestra sesion.
*/
class Response {
public:

  struct ResultCode {
    enum _v {
      Undefined,
      Success,
      Timeout, /**< answer timeout from peer, or transport failure with OTA request notification of timeout event deferred by configurtion */
      DiameterUnavailable /**< Diameter server was unavailable and notification is not deferred to timeout event for OTA requests */
    };
  };

  /**
     Devuelve el tipo de la clase de esta respuesta.
     \return El tipo de la clase de esta respuesta.
  */
  const ClassCode::_v & getClassCode() const throw() { return a_classCode; }

  /**
     Devuelve la identificacion del mensaje diameter.
     Esta identificacion sera generada automaticamente al enviar la peticion.
     \return Identificacion del mensaje asociado a esta peticion/respuesta.
     \see diameter::comm::Session::send
  */
  HopByHop getHopByHop() const throw() { return a_hopByHop; }

  /**
     Devuelve la sesion que genera esta respuesta.
     \return La instancia de la sesion que genera esta respuesta.
  */
  const Session* getSession() const throw() { return a_session; }

  /**
     Devuelve la sesion que origino la creacion de esta respuesta.
     \return La sesion que origino la creacion de esta respuesta.
  */
  Session* getSession() throw() { return a_session; }


  /**
     Devuelve el resultado de la peticion diameter solicitada.
     \return El resultado de la peticion diameter solicitada.
  */
  const ResultCode::_v & getResultCode() const throw() { return a_resultCode; }


  /**
     Returns original request for the response received from the network.
     \see diameter::comm::Session::send
  */
  const Message* getRequest() const throw() { return a_request; }

  /**
     Returns message received from the network.
     \see diameter::comm::Session::send
  */
  const anna::DataBlock *getMessage() const throw() { return a_message; }

  // helpers

//   /**
//      Reponse regarding diameter keepalive
//   */
//   bool isKeepAlive() const throw();

  /**
     Devuelve una cadena con la informacion relevante sobre esta instancia.
     \return Una cadena con la informacion relevante sobre esta instancia.
  */
  std::string asString() const throw();

private:
  ClassCode::_v a_classCode;
  HopByHop a_hopByHop;
  Session* a_session;
  ResultCode::_v a_resultCode;
  Timer* a_timer;
  const Message* a_request; // associated original request
  const anna::DataBlock *a_message; // received datablock from server

  typedef anna::Recycler<Response> response_pool;
  static response_pool st_responses;

  Response();

  static Response* instance(const ClassCode::_v &, const HopByHop) throw(anna::RuntimeException);
  static void release(Response* response) throw();

  void clear() throw();

  void setSession(Session* session) throw() { a_session = session; }
  void setRequest(const Message* request) throw() { a_request = request; }
  void setMessage(const anna::DataBlock *message) throw() { a_message = message; }
  void activateTimer() throw(anna::RuntimeException);
  void cancelTimer() throw();
  void setResultCode(const ResultCode::_v & resultCode) throw() { a_resultCode = resultCode; }

  static const char* asText(const ResultCode::_v) throw();

  friend class Session;
  friend class ClientSession;
  friend class ServerSession;
  friend class Timer;
  friend class anna::Allocator<Response>;
};

}
}
}

#endif

