// ANNA - Anna is Not Nothingness Anymore                                                         //
//                                                                                                //
// (c) Copyright 2005-2015 Eduardo Ramos Testillano & Francisco Ruiz Rayo                         //
//                                                                                                //
// See project site at http://redmine.teslayout.com/projects/anna-suite                           //
// See accompanying file LICENSE or copy at http://www.teslayout.com/projects/public/anna.LICENSE //


#ifndef anna_diameter_stack_Format_hpp
#define anna_diameter_stack_Format_hpp

// Local
#include <anna/diameter/codec/Format.hpp>

#include <anna/config/defines.hpp>
#include <anna/core/RuntimeException.hpp>

// STL
#include <string>



namespace anna {
namespace xml {
class Node;
}
}



namespace anna {

namespace diameter {

namespace stack {

class Dictionary;

//------------------------------------------------------------------------------
//----------------------------------------------------------------- class Format
//------------------------------------------------------------------------------
/**
* Format data container
*/
class Format {
  const Dictionary *a_dictionary;
  std::string a_name;
  std::string a_parentName;

  void _initialize(const Dictionary *d) throw() {
    a_dictionary = d;
    a_name = "";
    a_parentName = "";
  }

public:

  Format(const Dictionary *d = NULL) { _initialize(d); }
  ~Format() {};


  // get

  /**
  * Gets the format type name
  *
  * @return Format type name
  */
  const std::string & getName(void) const throw() { return a_name; }

  /**
  * Gets the format parent type name
  *
  * @return Format parent type name, empty string for diameter basic format types (and reserved like 'Any, Unknown')
  */
  const std::string & getParentName(void) const throw() { return a_parentName; }

  // helpers

  /**
  * Gets the diameter parent type reference
  *
  * @return Diameter parent type reference, NULL for diameter basic format types (and reserved like 'Any, Unknown')
  */
  const Format * getParent(void) const throw();

  /**
  * The diameter format is a derived type
  *
  * @return The format is derived
  */
  bool isDerived(void) const throw() { return (a_parentName != ""); }

  /**
  * The diameter format is a basic type:
  *  (OctetString, Integer32, Integer64, Unsigned32, Unsigned64, Float32, Float64, Grouped)
  *
  * @return The format is basic
  */
  bool isBasic(void) const throw() { return (!isDerived() && !isReserved()); }

  /**
  * Gets the diameter basic format type from which a derived type inherit, or basic type itself for non-derived
  *
  * @return Diameter basic format type
  */
  codec::Format::_v getBasicType(void) const throw(anna::RuntimeException);

  /**
  * The diameter format belongs to RFC3588 diameter format family:
  *  (OctetString, Integer32, Integer64, Unsigned32, Unsigned64, Float32, Float64, Grouped, Address, Time, UTF8String, DiameterIdentity, DiameterURI, Enumerated, IPFilterRule, QoSFilterRule)
  *
  * @return The format belongs to RFC3588
  */
  bool isRFC3588(void) const throw() { return (codec::Format::isRFC3588(a_name)); }

  // Reserved
  /** @return The format is 'Any' (generic AVP) */
  bool isAny() const throw() { return (a_name == codec::Format::asText(codec::Format::Any)); }
//   /** @return The format is 'Unknown' */
//   bool isUnknown() const throw() { return (a_name == codec::Format::asText(codec::Format::Unknown)); }

  // RFC 3588
  /** @return The format is 'OctetString' */
  bool isOctetString() const throw() { return (a_name == codec::Format::asText(codec::Format::OctetString)); }
  /** @return The format is 'Integer32' */
  bool isInteger32() const throw() { return (a_name == codec::Format::asText(codec::Format::Integer32)); }
  /** @return The format is 'Integer64' */
  bool isInteger64() const throw() { return (a_name == codec::Format::asText(codec::Format::Integer64)); }
  /** @return The format is 'Unsigned32' */
  bool isUnsigned32() const throw() { return (a_name == codec::Format::asText(codec::Format::Unsigned32)); }
  /** @return The format is 'Unsigned64' */
  bool isUnsigned64() const throw() { return (a_name == codec::Format::asText(codec::Format::Unsigned64)); }
  /** @return The format is 'Float32' */
  bool isFloat32() const throw() { return (a_name == codec::Format::asText(codec::Format::Float32)); }
  /** @return The format is 'Float64' */
  bool isFloat64() const throw() { return (a_name == codec::Format::asText(codec::Format::Float64)); }
  /** @return The format is 'Grouped' */
  bool isGrouped() const throw() { return (a_name == codec::Format::asText(codec::Format::Grouped)); }
  /** @return The format is 'Address' */
  bool isAddress() const throw() { return (a_name == codec::Format::asText(codec::Format::Address)); }
  /** @return The format is 'Time' */
  bool isTime() const throw() { return (a_name == codec::Format::asText(codec::Format::Time)); }
  /** @return The format is 'UTF8String' */
  bool isUTF8String() const throw() { return (a_name == codec::Format::asText(codec::Format::UTF8String)); }
  /** @return The format is 'DiameterIdentity' */
  bool isDiameterIdentity() const throw() { return (a_name == codec::Format::asText(codec::Format::DiameterIdentity)); }
  /** @return The format is 'DiameterURI' */
  bool isDiameterURI() const throw() { return (a_name == codec::Format::asText(codec::Format::DiameterURI)); }
  /** @return The format is 'Enumerated' */
  bool isEnumerated() const throw() { return (a_name == codec::Format::asText(codec::Format::Enumerated)); }
  /** @return The format is 'IPFilterRule' */
  bool isIPFilterRule() const throw() { return (a_name == codec::Format::asText(codec::Format::IPFilterRule)); }
  /** @return The format is 'QoSFilterRule' */
  bool isQoSFilterRule() const throw() { return (a_name == codec::Format::asText(codec::Format::QoSFilterRule)); }

  /**
  * The diameter format is application-specific or a format type name not registered on dictionary
  *
  * @return The format is application-specific
  */
  bool isApplicationSpecific(void) const throw() { return (!isRFC3588() && !isReserved()); }

  /**
  * The diameter format is reserved ('Any' for generic AVP, 'Unknown' for not registered avps)
  *
  * @return The format is reserved
  */
  bool isReserved(void) const throw() { return (codec::Format::isReserved(a_name)); }

  /**
  * Class string representation
  *
  * @return String with class content
  */
  std::string asString(void) const throw();

  /**
  * Class xml representation
  *
  * @return XML document with relevant information for this instance.
  */
  anna::xml::Node* asXML(anna::xml::Node* parent) const throw();

  // operators

  /**
  * @return Equality between two class instances
  */
  friend bool operator == (const Format & f1, const Format & f2) { return ((f1.getName() == f2.getName())); }

  /**
  * @return Difference between two class instances
  */
  friend bool operator != (const Format & f1, const Format & f2) { return !(f1 == f2); }

  // set

  /**
  * Initializes the class content
  */
  void initialize(const Dictionary *d = NULL) throw() { _initialize(d); }

  /**
  * Sets Avp format type name
  *
  * @param type Avp format type name
  */
  void setName(const char * name) throw(anna::RuntimeException) {
    if(name == NULL) throw anna::RuntimeException("Null Format-name not allowed", ANNA_FILE_LOCATION);

    a_name = name;

    if(a_name == "") throw anna::RuntimeException("Empty Format-name not allowed", ANNA_FILE_LOCATION);
  }

  /**
  * Sets RFC3588 Avp format type name
  *
  * @param rfc3588Format RFC3588 format type
  */
  void setRFC3588(codec::Format::_v rfc3588Format) throw() {
    setName(codec::Format::asText(rfc3588Format));
  }

  /**
  * Sets Diameter-basic avp format parent
  *
  * @param parent Diameter-basic avp format parent
  */
  void setParentName(const std::string & parentName) throw(anna::RuntimeException);
};


}
}
}


#endif


