// ANNA - Anna is Not Nothingness Anymore                                                         //
//                                                                                                //
// (c) Copyright 2005-2015 Eduardo Ramos Testillano & Francisco Ruiz Rayo                         //
//                                                                                                //
// See project site at http://redmine.teslayout.com/projects/anna-suite                           //
// See accompanying file LICENSE or copy at http://www.teslayout.com/projects/public/anna.LICENSE //


#ifndef anna_ldap_Response_hpp
#define anna_ldap_Response_hpp

#include <anna/core/util/Recycler.hpp>
#include <anna/core/Allocator.hpp>

#include <anna/ldap/defines.hpp>
#include <anna/ldap/ClassCode.hpp>
#include <anna/ldap/ResultCode.hpp>

namespace anna {

namespace ldap {

class Session;
class Attribute;
class Timer;
class Request;

/**
   Modela las respuesta que podemos recibir de un servidor LDAP.

   La respuesta correspondiente a una peticion se genera automaticamente al invocar a ldap::Session::send.
   El nucleo de ANNA.ldap notifica el resultado de la operacion solicitada mediante la invocacion
   al metodo-menejador ldap::Session::eventResponse de nuestra sesion.
*/
class Response {
public:
  typedef Recycler<Attribute> attribute_container;
  typedef attribute_container::iterator attribute_iterator;
  typedef attribute_container::const_iterator const_attribute_iterator;

  typedef Recycler<std::string> referral_container;
  typedef referral_container::const_iterator const_referral_iterator;

  /**
     Devuelve el tipo de la clase de esta respuesta.
     \return El tipo de la clase de esta respuesta.
  */
  ClassCode::_v getClassCode() const throw() { return a_classCode; }

  /**
     Devuelve la identificacion del mensaje LDAP.
     Esta identificacion sera generada automaticamente al enviar la peticion.
     \return Identificacion del mensaje asociado a esta peticion/respuesta.
     \see ldap::Session::send
  */
  IdMessage getIdMessage() const throw() { return a_idMessage; }

  /**
     Devuelve la sesion que genera esta respuesta.
     \return La instancia de la sesion que genera esta respuesta.
  */
  const Session* getSession() const throw() { return a_session; }

  /**
     Devuelve el resultado de la peticion LDAP solicitada.
     \return El resultado de la peticion LDAP solicitada.
  */
  const ResultCode& getResultCode() const throw() { return a_resultCode; }

  /**
     Devuelve el nombre del DN asociado a esta peticion.
     \return El nombre del DN asociado a esta peticion.
  */
  const std::string& getName() const throw() { return a_name; }

  /**
     Devuelve la peticion que origino la creacion de esta respuesta.
     \return La peticion que origino la creacion de esta respuesta.
     \see ldap::Session::send
  */
  const Request* getRequest()const  throw() { return a_request; }

  /**
     Devuelve la sesion que origino la creacion de esta respuesta.
     \return La sesion que origino la creacion de esta respuesta.
  */
  Session* getSession() throw() { return a_session; }

  /**
   * Devuelve el n de atributos que contiene la respuesta.
   * \return el n de atributos que contiene la respuesta.
   */
  int attribute_size() const throw() { return a_attributes.size(); }

  /**
     Devuelve el iterador al comienzo de la lista de atributos asociados a esta respuesta.
     \return El iterador al comienzo de la lista de atributos asociados a esta respuesta.
  */
  const_attribute_iterator attribute_begin() const throw() { return a_attributes.begin(); }

  /**
     Devuelve el iterador al final de la lista de atributos asociados a esta respuesta.
     \return El iterador al final de la lista de atributos asociados a esta respuesta.
  */
  const_attribute_iterator attribute_end() const throw() { return a_attributes.end(); }

  /**
   * Devuelve el n de referencias que contiene la respuesta.
   * \return el n de referencias que contiene la respuesta.
   */
  int referral_size() const throw() { return a_referrals.size(); }

  /**
     Devuelve el iterador al comienzo de la lista de referencias asociados a esta respuesta.
     \return El iterador al comienzo de la lista de referencias asociados a esta respuesta.
  */
  const_referral_iterator referral_begin() const throw() { return a_referrals.begin(); }

  /**
     Devuelve el iterador al final de la lista de referencias asociados a esta respuesta.
     \return El iterador al final de la lista de referencias asociados a esta respuesta.
  */
  const_referral_iterator referral_end() const throw() { return a_referrals.end(); }

  /**
     Devuelve el atributo asociado al iterador recibido como parametro.
     \param ii Instancia del iterador usado para recorrer los datos. Estara comprendido
     entre [#attribute_begin, #attribute_end).
     \return El atributo referenciado por el iterador.
  */
  static const Attribute* attribute(const_attribute_iterator ii) throw() { return attribute_container::data(ii); }

  /**
     Devuelve la referencia asociada al iterador recibido como parametro.
     \param ii Instancia del iterador usado para recorrer los datos. Estara comprendido
     entre [#referral_begin, #referral_end).
     \return La referencia apuntada por el iterador.
  */
  static const std::string& referral(const_referral_iterator ii) throw() { return *referral_container::data(ii); }

  /**
     Devuelve una cadena con la informacion relevante sobre esta instancia.
     \return Una cadena con la informacion relevante sobre esta instancia.
  */
  std::string asString() const throw();

private:
  ClassCode::_v a_classCode;
  IdMessage a_idMessage;
  Session* a_session;
  ResultCode a_resultCode;
  std::string a_name;
  attribute_container a_attributes;
  referral_container a_referrals;
  Timer* a_timer;
  const Request* a_request;

  typedef Recycler<Response> response_pool;
  static response_pool st_responses;

  Response();

  static Response* instance(const ClassCode::_v, const IdMessage) throw(RuntimeException);
  static void release(Response* response) throw();

  void clear() throw();

  void setName(const std::string& name) throw() { a_name = name; }
  void setSession(Session* session) throw() { a_session = session; }
  void setRequest(const Request* request) throw() { a_request = request; }
  void activateTimer() throw(RuntimeException);
  void cancelTimer() throw();
  void setResultCode(const ResultCode& resultCode) throw() { a_resultCode = resultCode; }

  Attribute* createAttribute(const std::string& name) throw(RuntimeException);

  void createReferral(const char* value) throw(RuntimeException) {
    std::string* newReferral = a_referrals.create();
    *newReferral = value;
  }


  attribute_iterator attribute_begin() throw() { return a_attributes.begin(); }
  attribute_iterator attribute_end() throw() { return a_attributes.end(); }
  static Attribute* attribute(attribute_iterator ii) throw() { return attribute_container::data(ii); }

  friend class Session;
  friend class Allocator<Response>;
};

}
}

#endif

