// ANNA - Anna is Not Nothingness Anymore                                                         //
//                                                                                                //
// (c) Copyright 2005-2015 Eduardo Ramos Testillano & Francisco Ruiz Rayo                         //
//                                                                                                //
// See project site at http://redmine.teslayout.com/projects/anna-suite                           //
// See accompanying file LICENSE or copy at http://www.teslayout.com/projects/public/anna.LICENSE //


#include <anna/core/tracing/Logger.hpp>
#include <anna/core/functions.hpp>
#include <anna/core/tracing/TraceMethod.hpp>

#include <anna/xml/Node.hpp>
#include <anna/xml/Attribute.hpp>

#include <anna/comm/ServerSocket.hpp>
#include <anna/comm/Communicator.hpp>
#include <anna/comm/internal/LocalConnection.hpp>
#include <anna/comm/ClientSocket.hpp>
#include <anna/comm/Device.hpp>

#include <anna/comm/handler/ServerSocket.hpp>
#include <anna/comm/handler/Manager.hpp>

using namespace std;
using namespace anna;

void comm::handler::ServerSocket::initialize()
throw(RuntimeException) {
  LOGMETHOD(TraceMethod traceMethod("comm::handler::ServerSocket", "initialize", ANNA_FILE_LOCATION));

  if(a_serverSocket == NULL) {
    string msg(asString());
    msg += " | comm::ServerSocket was not established";
    throw RuntimeException(msg, ANNA_FILE_LOCATION);
  }

  if(a_serverSocket->isBound() == false)
    a_serverSocket->bind();

  setfd(a_serverSocket->getfd());
  a_serverSocket->prepare();
  a_serverSocket->setBlockingMode(false);
}

/*
// Cuando un ServerSocket recibe actividad en el fd slo puede ser para
// recibir una solicitud de conexin de un proceso remoto, que estar
// instanciando un ClientSocket contra este.
//
// (1) Esta situacion se ha dado cuando hemos probado varios servidores
// en modo compartido.
//
// Se invoca desde [Tx] -> <null>
// La secuencia de bloqueo tiene que coincidir con la del mtodo comm::handler::LocalConnection::finalize:

<Method Id="comm::handler::MetaClientSocket::apply" File="handler.db/comm.handler.MetaClientSocket.cc" Line="45" LineNo="281">
    <Guard Id="0x68ccd0" Name="comm::ClientSocket from comm::handler::MetaClientSocket::apply" LineNo="283"/>
    <Guard Id="0x68c6a0" Name="comm::Communicator::detach" LineNo="289">
       <Method Id="comm::handler::LocationConnection::finalize" File="handler.db/comm.handler.LocalConnection.cc" Line="78" LineNo="293">
         <Guard Id="0x68ccd0" Name="comm::Socket::close" LineNo="295">
            <Guard Id="0x6818e0" Name="PN7anna4comm16TransportFactoryE" LineNo="297"/>
         </Guard>
         <Guard Id="0x68c678" Name="N7anna12SafeRecyclerINS_4comm15LocalConnectionENS_9AllocatorIS2_EEEE::release (T*)" LineNo="305"/>
        </Method>
    </Guard>
</Method>

// para eso bloqueamos el comm::Communicator.
*/
void comm::handler::ServerSocket::apply()
throw(RuntimeException) {
  LOGMETHOD(TraceMethod traceMethod("handler::ServerSocket", "apply", ANNA_FILE_LOCATION));
  Guard guard(a_communicator, "Communicator from handler::ServerSocket");
  comm::LocalConnection* localConnection = a_serverSocket->accept();

  if(localConnection == NULL)                                          // (1)
    return;

  bool accept = a_communicator->eventAcceptConnection(*localConnection->getClientSocket());

  // @Eduardo (independence from communicator)
  if(accept) accept = a_serverSocket->eventAcceptConnection(*localConnection->getClientSocket());

  // Ver comentarios adicionales en void comm::handler::LocalConnection::finalize ()

  if(accept == false) {
    LOGWARNING(
      string msg(localConnection->asString());
      msg += " | Service denied";
      Logger::warning(msg, ANNA_FILE_LOCATION);
    );
    localConnection->getClientSocket()->close();
    a_serverSocket->release(localConnection);
  } else {
    try {
      a_communicator->attach(localConnection);
    } catch(RuntimeException&) {
      localConnection->getClientSocket()->close();
      a_serverSocket->release(localConnection);
      throw;
    }
  }
}

//--------------------------------------------------------------------------
// Si cae la direccin en la que est escuchando finalizamos todos los
// que tengan establecidos contra este ServerSocket.
//--------------------------------------------------------------------------
void comm::handler::ServerSocket::breakAddress(const in_addr_t& address)
throw() {
  LOGMETHOD(TraceMethod traceMethod("comm::handler::ServerSocket", "breakAddress", ANNA_FILE_LOCATION));
  const comm::AccessPoint& accessPoint = a_serverSocket->getLocalAccessPoint();
  const comm::Device* device = accessPoint.getINetAddress().getDevice(false);

  if(device == NULL)
    return;

  if(device->getAddress() != address)
    return;

  Guard guard(a_serverSocket, "comm::handler::ServerSocket::breakAddress");
  /**
   * Trabaja sobre una copia para no perder la referencia cuando se elimine un miembro de la lista original
   */
  typedef vector <comm::LocalConnection*> work_container;
  typedef work_container::iterator work_iterator;
  work_container ww;

  for(comm::ServerSocket::iterator ii = a_serverSocket->begin(), maxii = a_serverSocket->end(); ii != maxii; ii ++)
    ww.push_back(comm::ServerSocket::localConnection(ii));

  int n = 0;

  for(work_iterator ii = ww.begin(), maxii = ww.end(); ii != maxii; ii ++) {
    a_communicator->detach((*ii)->getClientSocket());
    n ++;
  }

  LOGDEBUG(
    string msg("comm::handler::ServerSocket::breakAddress | ");
    msg += a_serverSocket->asString();
    msg += " | Clients number: ";
    msg += functions::asString(n);
    Logger::debug(msg, ANNA_FILE_LOCATION);
  );
}

/**
 * No cierra las conexiones locales que hayan partido desde este ServerSocket, porque tambien cerriamos el
 * MainHandler cuando se use el Communicator::Clone.
 */
void comm::handler::ServerSocket::finalize()
throw() {
  LOGMETHOD(TraceMethod traceMethod("comm::handler::ServerSocket", "finalize", ANNA_FILE_LOCATION));

  if(a_serverSocket == NULL)
    return;

  for(comm::ServerSocket::iterator ii = a_serverSocket->begin(), maxii = a_serverSocket->end(); ii != maxii; ii ++)
    comm::ServerSocket::localConnection(ii)->setServerSocket(NULL);

  a_serverSocket->close();
  a_serverSocket = NULL;
}

string comm::handler::ServerSocket::asString() const
throw() {
  string result("comm::handler::ServerSocket { ");
  result += comm::Handler::asString();
  result += " | ";
  result += functions::asString(a_serverSocket);
  return result += " }";
}

xml::Node* comm::handler::ServerSocket::asXML(xml::Node* parent) const
throw() {
  xml::Node* result = parent->createChild("comm.handler.ServerSocket");
  comm::Handler::asAttribute(result);

  if(a_serverSocket)
    a_serverSocket->asXML(result);

  return result;
}


