// ANNA - Anna is Not Nothingness Anymore                                                         //
//                                                                                                //
// (c) Copyright 2005-2015 Eduardo Ramos Testillano & Francisco Ruiz Rayo                         //
//                                                                                                //
// See project site at http://redmine.teslayout.com/projects/anna-suite                           //
// See accompanying file LICENSE or copy at http://www.teslayout.com/projects/public/anna.LICENSE //


#include <stdio.h>
#include <ctype.h>

#include <algorithm>

#include <anna/core/functions.hpp>
#include <anna/core/util/Configuration.hpp>
#include <anna/core/RuntimeException.hpp>
#include <anna/core/tracing/Logger.hpp>
#include <anna/core/tracing/TraceMethod.hpp>

using namespace std;
using namespace anna;

const char* Configuration::defaultSection = "@global@";

void Configuration::load(const char* configFile)
throw(RuntimeException) {
  LOGMETHOD(TraceMethod tm("Configuration", "load", ANNA_FILE_LOCATION));
  char buffer [2048];
  FILE* file;
  char* aux;
  string currentSection(defaultSection);
  int nline(1);
  removeAll();

  if((file = fopen(configFile, "r")) == NULL)
    throw RuntimeException(configFile, errno, ANNA_FILE_LOCATION);

  LOGDEBUG(Logger::write(Logger::Debug, "Configuration file", configFile, ANNA_FILE_LOCATION));

  try {
    while(fgets(buffer, sizeof(buffer) - 1, file) != NULL) {
      if((aux = strchr(buffer, '#')) != NULL)
        * aux = 0;

      if(anna_strlen(aux = strip(buffer)) <= 0)
        continue;

      if(processSection(nline, aux, currentSection) == true)
        continue;

      processVariable(nline, aux, currentSection);
      nline ++;
    }
  } catch(RuntimeException& ex) {
    fclose(file);
    throw;
  }

  fclose(file);
}

bool Configuration::exists(const char* sectionName, const char* variableName) const
throw() {
  const VariableEx* var = find(string(sectionName), variableName);
  return (var == NULL) ? false : !var->isNull();
}

void Configuration::setDefaultValue(const char* sectionName, const char* variableName, const char* defaultValue)
throw(RuntimeException) {
  string section(sectionName);
  VariableEx* var = find(section, variableName);

  if(var == NULL)
    var = createVariable(section, variableName);

  var->setDefaultValue(defaultValue);
  LOGINFORMATION(
    string msg(sectionName);
    msg += "::";
    msg += variableName;
    msg += " (default)";
    msg += " | Default value: ";
    msg += defaultValue;
    Logger::write(Logger::Information, msg, ANNA_FILE_LOCATION);
  )
}

const char* Configuration::getValue(const char* sectionName, const char* variableName, const bool strict) const
throw(RuntimeException) {
  const VariableEx* variable = find(string(sectionName), variableName);
  const char* result(NULL);

  if(variable == NULL) {
    string msg("Variable ");
    msg += sectionName;
    msg += "::";
    msg += variableName;
    msg += " is not defined";
    throw RuntimeException(msg, ANNA_FILE_LOCATION);
  }

  if(variable->isNull() == false)
    result = variable->getStringValue();
  else
    result = (strict == false) ? variable->getDefaultValue() : NULL;

  return result;
}

int Configuration::getIntegerValue(const char* sectionName, const char* variableName, const bool strict) const
throw(RuntimeException) {
  return atoi(getValue(sectionName, variableName, strict));
}

char* Configuration::strip(char* buffer) {
  char* result;
  result = buffer + (anna_strlen(buffer) - 1);

  while(result >= buffer && isspace(*result))
    result --;

  result[1] = 0;

  for(result = buffer; *result  && isspace(*result); result ++);

  return result;
}

bool Configuration::processSection(const int nline, char* buffer, string& currentSection) {
  bool result(false);
  char* end;
  char* section;

  if(*buffer == '[') {
    if((end = strchr(buffer, ']')) != NULL) {
      *end = 0; // JASC, antes NULL

      if(anna_strlen(section = strip(++ buffer)) > 0)
        currentSection = section;
      else {
        string msg("Invalid section name at line: ");
        msg += functions::asString(nline);
        throw RuntimeException(msg, ANNA_FILE_LOCATION);
      }

      result = true;
    }
  }

  return result;
}

void Configuration::processVariable(const int nline, char* buffer, const string& currentSection)
throw(RuntimeException) {
  char* variableName;
  char* value;
  char* aux;
  VariableEx* variable;

  if((aux = anna_strchr(buffer , '=')) == NULL)
    return;

  *aux = 0;

  if(anna_strlen(variableName = strip(buffer)) <= 0) {
    string msg("Invalid section name at line: ");
    msg += functions::asString(nline);
    throw RuntimeException(msg, ANNA_FILE_LOCATION);
  }

  if(anna_strlen(value = strip(aux + 1)) <= 0)
    return;

  if((variable = find(currentSection, variableName)) != NULL) {
    if(variable->isNull() == false) {
      string msg("Duplicated variable | Section: ");
      msg += currentSection;
      msg += " | Variable: ";
      msg += variableName;
      msg += " | Line: ";
      msg += functions::asString(nline);
      throw RuntimeException(msg, ANNA_FILE_LOCATION);
    }
  } else
    variable = createVariable(currentSection, variableName);

  variable->setValue(value);

  if(Logger::isActive(Logger::Information)) {
    string msg(currentSection);
    msg += "::";
    msg += variableName;
    msg += " | Value: ";
    msg += value;
    Logger::write(Logger::Information, variable->asString(), ANNA_FILE_LOCATION);
  }
}

Configuration::VariableEx* Configuration::createVariable(const string& section, const char* variableName)
throw() {
  map <string, VariableEx::Vector*>::iterator isection(a_sections.find(section));
  VariableEx* result;
  VariableEx::Vector* variables;
  result = new VariableEx(variableName);

  if(isection == a_sections.end()) {
    variables = new VariableEx::Vector;
    a_sections [section] = variables;
  } else
    variables = isection->second;

  variables->push_back(result);
  return result;
}

Configuration::VariableEx* Configuration::find(const string& section, const char* variableName)
throw() {
  VariableEx* result(NULL);
  map <string, VariableEx::Vector*>::iterator isection(a_sections.find(section));

  if(isection == a_sections.end())
    return NULL;

  VariableEx::Vector* varVector(isection->second);

  for(VariableEx::Vector::iterator vv = varVector->begin(), maxvv = varVector->end(); vv != maxvv; vv ++) {
    if(!strcmp((*vv)->getName(), variableName)) {
      result = *vv;
      break;
    }
  }

  return result;
}

void Configuration::removeAll()
throw() {
  map <string, VariableEx::Vector*>::iterator ii(a_sections.begin());
  map <string, VariableEx::Vector*>::iterator end(a_sections.end());
  VariableEx::Vector::iterator vv, maxvv;
  VariableEx* variable;

  for(; ii != end; ii ++) {
    for(vv = ii->second->begin(), maxvv = ii->second->end(); vv != maxvv; vv ++) {
      variable = *vv;
      delete variable;
    }

    delete ii->second;
  }

  a_sections.clear();
}

