// ANNA - Anna is Not Nothingness Anymore                                                         //
//                                                                                                //
// (c) Copyright 2005-2015 Eduardo Ramos Testillano & Francisco Ruiz Rayo                         //
//                                                                                                //
// See project site at http://redmine.teslayout.com/projects/anna-suite                           //
// See accompanying file LICENSE or copy at http://www.teslayout.com/projects/public/anna.LICENSE //


#include <anna/core/functions.hpp>
#include <anna/diameter/defines.hpp>
#include <anna/diameter/functions.hpp>
#include <anna/diameter/helpers/helpers.hpp>
#include <anna/diameter/codec/functions.hpp>
#include <anna/diameter/codec/Message.hpp>
#include <anna/diameter/helpers/base/functions.hpp>
#include <anna/time/functions.hpp>

#include <anna/diameter.comm/ClientSession.hpp>
#include <anna/diameter.comm/Engine.hpp>
#include <anna/diameter.comm/Entity.hpp>
#include <anna/diameter.comm/Server.hpp>
#include <anna/diameter.comm/Response.hpp>
#include <anna/diameter.comm/Message.hpp>
#include <anna/diameter.comm/OamModule.hpp>
#include <anna/diameter.comm/ApplicationMessageOamModule.hpp>
#include <anna/diameter.comm/TimerManager.hpp>
#include <anna/diameter.comm/Timer.hpp>
#include <anna/diameter.comm/ClientSessionReceiver.hpp>
#include <anna/diameter.comm/ReceiverFactoryImpl.hpp>

#include <anna/comm/ClientSocket.hpp>
#include <anna/app/functions.hpp>
#include <anna/core/functions.hpp>
#include <anna/core/DataBlock.hpp>
#include <anna/core/tracing/Logger.hpp>
#include <anna/core/tracing/TraceMethod.hpp>
#include <anna/xml/Node.hpp>
#include <anna/timex/Engine.hpp>

// Standard
#include <stdlib.h> // rand()
#include <time.h>


using namespace std;
using namespace anna::diameter;
using namespace anna::diameter::comm;

//static
const anna::Millisecond ClientSession::DefaultWatchdogPeriod(30000); // Watchdog messages timeout


ClientSession::ClientSession() : Session("diameter::comm::ClientSession", "Diameter Keep-Alive Timer"),
  a_receiverFactory(this),
  a_cer(ClassCode::Bind),
  a_dwr(ClassCode::ApplicationMessage) // realmente no es necesario, los Message son por defecto de aplicacion
{ initialize(); }


void ClientSession::initialize() throw() {
  Session::initialize();
  a_autoRecovery = true;
  a_parent = NULL;
  a_server = NULL;
  a_watchdogState = WatchdogState::TimerStopped;
  a_hidden = false;
}

//ClientSession::~ClientSession() {;}


const std::string& ClientSession::getAddress() const throw() {
  return a_parent->getAddress();
}

int ClientSession::getPort() const throw() {
  return a_parent->getPort();
}


void ClientSession::setState(State::_v state) throw() {
  Session::setState(state);
  // Inform father server (availability changes):
  a_parent->refreshAvailability();
}

void ClientSession::bind() throw(anna::RuntimeException) {
  LOGMETHOD(anna::TraceMethod traceMethod(a_className, "bind", ANNA_FILE_LOCATION));

  if(a_state != State::Closed)
    return;

  LOGDEBUG(
    string msg("diameter::comm::ClientSession::bind | ");
    msg += asString();
    anna::Logger::debug(msg, ANNA_FILE_LOCATION);
  );

  // Connection:
  if(!a_server)
    throw anna::RuntimeException("Server is not yet created", ANNA_FILE_LOCATION);

  bool serverAvailable = a_server->isAvailable();
  LOGDEBUG(

    if(serverAvailable) anna::Logger::debug("Server AVAILABLE", ANNA_FILE_LOCATION);
    else anna::Logger::debug("Server NOT AVAILABLE. Connecting ...", ANNA_FILE_LOCATION);
    );

  if(!serverAvailable) {
    a_server->connect();
    return;
  }

  // Some operations could be done before sending CER, for example non-standard Origin-Host manipulation for
  // Tekelec PCRF
  a_engine->bindingClientSession(this);

  // OAM Lo comento, porque no se contabilizan los reintentos y por lo tanto no son muy �tiles.
//  OamModule &oamModule = OamModule::instantiate();
//  oamModule.count(a_server->isAvailable() ? OamModule::Counter::TCPConnectOK:OamModule::Counter::TCPConnectNOK);
  // Application bind
  if(send(&a_cer))
    LOGDEBUG(anna::Logger::debug("CER sent to the server", ANNA_FILE_LOCATION));
}


void ClientSession::setCERandDWR(const anna::DataBlock & cer, const anna::DataBlock & dwr) throw(anna::RuntimeException) {
  if(codec::functions::getCommandId(cer) != helpers::base::COMMANDID__Capabilities_Exchange_Request) {
    throw anna::RuntimeException("The message provided as 'CER' is not a Capabilities-Exchange-Request", ANNA_FILE_LOCATION);
  }

  if(codec::functions::getCommandId(dwr) != helpers::base::COMMANDID__Device_Watchdog_Request) {
    throw anna::RuntimeException("The message provided as 'DWR' is not a Device-Watchdog-Request", ANNA_FILE_LOCATION);
  }

  // La verificacion ya se hace implicitamente antes
  //   if ((a_cer.isEmpty()) || (a_dwr.isEmpty())) {
  //      LOGDEBUG (anna::Logger::debug ("Must define valid CER and DWR messages before use bind !", ANNA_FILE_LOCATION));
  //      return;
  //   }
  a_cer.setBody(cer);
  a_dwr.setBody(dwr);
}


const Response* ClientSession::send(const Message* message) throw(anna::RuntimeException) {
  LOGMETHOD(anna::TraceMethod traceMethod(a_className, "send", ANNA_FILE_LOCATION));

  if(!message)
    throw anna::RuntimeException("Cannot send a NULL message", ANNA_FILE_LOCATION);

  // Command id:
  bool isRequest;
  diameter::CommandId cid = message->getCommandId(isRequest);
  diameter::ApplicationId aid = message->getApplicationId();

  LOGDEBUG(
    std::string msg = "Sending diameter message: ";
    msg += anna::diameter::functions::commandIdAsPairString(cid);
    anna::Logger::debug(msg, ANNA_FILE_LOCATION);
  );

  if(hidden()) {
    if((cid.first != helpers::base::COMMANDID__Capabilities_Exchange_Request.first)  /* not CER/CEA */
        && (cid.first != helpers::base::COMMANDID__Device_Watchdog_Request.first) /* not DWR/DWA */
        && (cid.first != helpers::base::COMMANDID__Disconnect_Peer_Request.first)) { /* not DPR/DPA */
      LOGDEBUG(
        std::string msg(asString());
        msg += " | Client-session hidden for application messages delivery";
        anna::Logger::debug(msg, ANNA_FILE_LOCATION);
      );
      return NULL;
    }
  }

  // Checkings:
  if((a_state == State::Closed) && (cid != helpers::base::COMMANDID__Capabilities_Exchange_Request)) {
    string msg(asString());
    msg += " | ClientSession::bind is not initiated";
    throw anna::RuntimeException(msg, ANNA_FILE_LOCATION);
  }

  if(a_state == State::WaitingBind) {
    string msg(asString());
    msg += " | Still waiting for bind ack (CEA)";
    throw anna::RuntimeException(msg, ANNA_FILE_LOCATION);
  }

  if(a_state == State::Failover) {
    string msg(asString());
    msg += " | Disabled sent on failover state";
    throw anna::RuntimeException(msg, ANNA_FILE_LOCATION);
  }

  if(a_state == State::Closing) {
    string msg(asString());
    msg += " | Disabled sent on closing state";
    throw anna::RuntimeException(msg, ANNA_FILE_LOCATION);
  }

  // Check states:
  if(cid == helpers::base::COMMANDID__Capabilities_Exchange_Request) {
    if(a_state != State::Closed) {
      string msg(asString());
      msg += " | Discarding CER on not closed state";
      throw anna::RuntimeException(msg, ANNA_FILE_LOCATION);
    }
  } else if(cid ==  helpers::base::COMMANDID__Device_Watchdog_Request) {
    if(a_state == State::WaitingDPA) {
      string msg(asString());
      msg += " | DWR is not sent on 'WaitingDPA' state";
      //throw anna::RuntimeException(msg, ANNA_FILE_LOCATION);
      LOGDEBUG(anna::Logger::debug(msg, ANNA_FILE_LOCATION));
      return NULL;
    }

    if(a_state == State::Disconnecting) {
      string msg(asString());
      msg += " | DWR is not sent on 'Disconnecting' state";
      //throw anna::RuntimeException(msg, ANNA_FILE_LOCATION);
      LOGDEBUG(anna::Logger::debug(msg, ANNA_FILE_LOCATION));
      return NULL;
    }

    if(a_state == State::Closing) {
      string msg(asString());
      msg += " | DWR is not sent on 'Closing' state";
      //throw anna::RuntimeException(msg, ANNA_FILE_LOCATION);
      LOGDEBUG(anna::Logger::debug(msg, ANNA_FILE_LOCATION));
      return NULL;
    }
  } else if(cid ==  helpers::base::COMMANDID__Disconnect_Peer_Request) {
    if(a_state == State::WaitingDPA) {
      string msg(asString());
      msg += " | Still waiting for DPR ack (DPA)";
      throw anna::RuntimeException(msg, ANNA_FILE_LOCATION);
    }

    if(a_state == State::Disconnecting) {
      string msg(asString());
      msg += " | Client disconnection has already been initiated";
      throw anna::RuntimeException(msg, ANNA_FILE_LOCATION);
    }
  } else {
    if((a_state == State::WaitingDPA) || (a_state == State::Disconnecting)) {
      if(cid != helpers::base::COMMANDID__Disconnect_Peer_Answer) {
        LOGDEBUG(
          string msg("diameter::comm::ClientSession::send | ");
          msg += asString();
          msg += " | Sents (request or answer) blocked to diameter server (disconnection in progress). Discarding ...";
          anna::Logger::debug(msg, ANNA_FILE_LOCATION);
        );
        return NULL;
      }
    }
  }

  // Trace send operation:
  LOGDEBUG(
    string msg("diameter::comm::ClientSession::send | ");
    msg += asString();
    msg += " | ";
    msg += message->asString();
    anna::Logger::debug(msg, ANNA_FILE_LOCATION);
  );
  bool fixed = false; // answers cannot be fixed
  Message * message_nc = const_cast<Message*>(message);

  if(isRequest) {
    if(/* entity */getParent()->getParent()->isDeprecated()) {
      string msg(asString());
      msg += " | Parent entity is deprecated. Request send blocked.";
      throw anna::RuntimeException(msg, ANNA_FILE_LOCATION);
    }

    // Fixing indicator:
    fixed = message_nc->fixRequestSequence(a_nextHopByHop, a_nextEndToEnd);
    message_nc->updateRequestTimestampMs(); // statistics purposes (processing time for request type)
  }

  // Send message
  try {
    message->send(*this);

    // Next hop by hop & end to end identifiers:
    if(isRequest) generateNextSequences();

    //   Transaction state
    //         The Diameter protocol requires that agents maintain transaction
    //         state, which is used for failover purposes.  Transaction state
    //         implies that upon forwarding a request, the Hop-by-Hop identifier
    //         is saved; the field is replaced with a locally unique identifier,
    //         which is restored to its original value when the corresponding
    //         answer is received.  The request's state is released upon receipt
    //         of the answer.  A stateless agent is one that only maintains
    //         transaction state.
    //
    updateOutgoingActivityTime();
    // OAM
    countSendings(cid, aid, true /* send ok */);
    // Trace non-application messages:
    LOGDEBUG(
      if((cid == helpers::base::COMMANDID__Capabilities_Exchange_Request) ||
         (cid == helpers::base::COMMANDID__Device_Watchdog_Request) ||
         (cid == helpers::base::COMMANDID__Disconnect_Peer_Request)) {
      anna::Logger::debug("Sent DataBlock to XML representation:", ANNA_FILE_LOCATION);
      try {
        anna::diameter::codec::Message msg(a_engine->getBaseProtocolCodecEngine()); msg.decode(message->getBody()); /* decode to be traced */
      }
      catch(anna::RuntimeException &ex) {
        std::string msg = ex.getText();
        msg += " | Use diameter::comm::Engine::setBaseProtocolCodecEngine() to allow internal base protocol messages full tracing";
        anna::Logger::debug(msg, ANNA_FILE_LOCATION);
      }
    }
    );

    // Restore sequences:
    if(fixed) message_nc->restoreSequencesAfterFix();  // restore to application sequences after fix
  } catch(anna::RuntimeException&) {
    if(fixed) message_nc->restoreSequencesAfterFix();  // restore to application sequences after fix

    // OAM
    countSendings(cid, aid, false /* send no ok */);
    throw;
  }

  // Renew states:
  if(cid ==  helpers::base::COMMANDID__Capabilities_Exchange_Request) {
    setState(State::WaitingBind);
  } else if(cid ==  helpers::base::COMMANDID__Disconnect_Peer_Request) {
    LOGWARNING(anna::Logger::warning("DPR has been sent to the peer (diameter server)", ANNA_FILE_LOCATION));
    setState(State::WaitingDPA);
  }

//   else {
//      // No changes
//   }

  // Answers are not temporized:
  if(!isRequest) return NULL;

  // Neither DWR:
  if(cid == helpers::base::COMMANDID__Device_Watchdog_Request) {
    setWatchdogState(WatchdogState::WaitingDWA);
    return NULL;
  }

  // Request will have context responses:
  Response* result(NULL);
  result = Response::instance(message->getClassCode(), a_nextHopByHop - 1 /* current request sent to server */);
  result->setRequest(message);
  response_add(result);
  return result;
}

bool ClientSession::unbind(bool forceDisconnect)
throw(anna::RuntimeException) {
  LOGMETHOD(anna::TraceMethod traceMethod(a_className, "unbind", ANNA_FILE_LOCATION));

  if(a_state == State::Closed)
    return false;

  // Client socket:
  anna::comm::ClientSocket * cs = const_cast<anna::comm::ClientSocket*>(a_server->getClientSocket());
//   LOGDEBUG(
//      string msg("Server to be unbound | ");
//      msg += a_server->asString();
//      anna::Logger::debug(msg, ANNA_FILE_LOCATION);
//   );

  if(forceDisconnect) {
    LOGDEBUG(anna::Logger::debug("Immediate disconnection (forceDisconnect)", ANNA_FILE_LOCATION));

    if(cs) cs->requestClose();  // this will invoke finalize()

    return true;
  }

  if(a_state == State::Disconnecting) {
    LOGDEBUG(
      string msg("diameter::comm::ClientSession::unbind | ");
      msg += asString();
      msg += " | Disconnection already in progress !";
      anna::Logger::debug(msg, ANNA_FILE_LOCATION);
    );
    return false;
  }

  if(a_state == State::Failover) {
    LOGDEBUG(
      string msg("diameter::comm::ClientSession::unbind | ");
      msg += asString();
      msg += " | Unbind on failover state. Disconnect now.";
      anna::Logger::debug(msg, ANNA_FILE_LOCATION);
    );

    if(cs) cs->requestClose();  // this will invoke finalize()

    return true;
  }

  if(a_state == State::WaitingBind) {
    LOGDEBUG(
      string msg("diameter::comm::ClientSession::unbind | ");
      msg += asString();
      msg += " | Unbind on WaitingBind state. Disconnect now.";
      anna::Logger::debug(msg, ANNA_FILE_LOCATION);
    );

    if(cs) cs->requestClose();  // this will invoke finalize()

    return true;
  }

  if(a_onDisconnect == OnDisconnect::IgnorePendings) {
    LOGDEBUG(anna::Logger::debug("Immediate disconnection (IgnorePendings)", ANNA_FILE_LOCATION));

    if(cs) cs->requestClose();  // this will invoke finalize()

    return true;
  }

  if(getOTARequests() == 0) {  // No pendings
    LOGDEBUG(anna::Logger::debug("No pending answers. Perform client-session close.", ANNA_FILE_LOCATION));

    if(cs) cs->requestClose();  // this will invoke finalize()

    return true;
  }

  if(a_state == State::Closing) {
    LOGDEBUG(
      string msg("diameter::comm::ClientSession::unbind | ");
      msg += asString();
      msg += " | Closing already in progress (waiting pendings) !";
      anna::Logger::debug(msg, ANNA_FILE_LOCATION);
    );
  }

  return false;
}

void ClientSession::eventPeerShutdown() throw() {
  // Inform father server:
  a_parent->eventPeerShutdown(this);
}

void ClientSession::eventRequestRetransmission(Message *request) throw() {

  // OAM
  OamModule &oamModule = OamModule::instantiate();
  oamModule.count(OamModule::Counter::RequestRetransmitted);
  oamModule.count(OamModule::Counter::RequestRetransmittedOnClientSession);

  // Inform father server:
  a_parent->eventRequestRetransmission(this, request);
}

void ClientSession::eventResponse(const Response& response) throw(anna::RuntimeException) {
  // Inform father server:
  a_parent->eventResponse(response);
}

void ClientSession::eventRequest(const anna::DataBlock &request) throw(anna::RuntimeException) {
  // Inform father server:
  a_parent->eventRequest(this, request);
}

void ClientSession::eventUnknownResponse(const anna::DataBlock& response) throw(anna::RuntimeException) {
  // Inform father server:
  a_parent->eventUnknownResponse(this, response);
}

void ClientSession::eventDPA(const anna::DataBlock& response) throw(anna::RuntimeException) {
  // Inform father server:
  a_parent->eventDPA(this, response);
}



//------------------------------------------------------------------------------------------
// Se invoca desde el diameter::comm::Receiver
//------------------------------------------------------------------------------------------
void ClientSession::receive(const anna::comm::Message& message)
throw(anna::RuntimeException) {
  LOGMETHOD(anna::TraceMethod traceMethod(a_className, "receive", ANNA_FILE_LOCATION));
  // Activity:
  updateIncomingActivityTime();
  activateTimer();
  // Command id:
  const anna::DataBlock & db = message.getBody();
  diameter::CommandId cid = codec::functions::getCommandId(db);
  bool isRequest = cid.second;
  LOGDEBUG(
    std::string msg = "Received diameter message: ";
    msg += anna::diameter::functions::commandIdAsPairString(cid);
    anna::Logger::debug(msg, ANNA_FILE_LOCATION);

    if( (cid == helpers::base::COMMANDID__Capabilities_Exchange_Answer) ||
        (cid.first == helpers::base::COMMANDID__Device_Watchdog_Request.first)) {
      try {
        anna::diameter::codec::Message dmsg(a_engine->getBaseProtocolCodecEngine()); dmsg.decode(db); /* decode to be traced */
      }
      catch(anna::RuntimeException &ex) {
        std::string msg = ex.getText();
        msg += " | Use diameter::comm::Engine::setBaseProtocolCodecEngine() to allow internal base protocol messages full tracing";
        anna::Logger::debug(msg, ANNA_FILE_LOCATION);
      }
    }
);

  // Main counters:
  OamModule &oamModule = OamModule::instantiate();
  oamModule.count(isRequest ? OamModule::Counter::RequestReceived : OamModule::Counter::AnswerReceived);
  oamModule.count(isRequest ? OamModule::Counter::RequestReceivedOnClientSession : OamModule::Counter::AnswerReceivedOnClientSession);
  // Statistic (size)
  a_parent->updateReceivedMessageSizeStatisticConcept(message.getSize(), cid); // only on reception (application could manage sent sizes)

  if(isRequest) {
    /////////////////////////////
    // Here received a request //
    /////////////////////////////

    // Received CER
    if(cid == helpers::base::COMMANDID__Capabilities_Exchange_Request) {
      LOGWARNING(anna::Logger::warning("Received CER: unexpected message at client-side", ANNA_FILE_LOCATION));
      return;
    }
    // Received DWR
    else if(cid == helpers::base::COMMANDID__Device_Watchdog_Request) {
//         LOGWARNING(anna::Logger::warning("Received DWR: unexpected message at client-side", ANNA_FILE_LOCATION));
//         return;
      // Non-usual but could happen:
      oamModule.count(OamModule::Counter::DWRReceived);
      sendDWAToServer(db /* DWR datablock received from server */);
      return;
    }
    // Received DPR
    else if(cid == helpers::base::COMMANDID__Disconnect_Peer_Request) {
      oamModule.count(OamModule::Counter::DPRReceived);

      if(a_state == State::Bound) {
        a_dpr.setBody(db);
        setState(State::Disconnecting);
        LOGWARNING(anna::Logger::warning("DPR has been received from peer (diameter server)", ANNA_FILE_LOCATION));

        if(getOTARequests() == 0) sendDPA();

        return; // DPR won't be informed because virtual readDPA is available for this
      }
    }

    try {
      // application message counters
      ApplicationMessageOamModule::instantiate().count(cid.first, -1 /* no result code */, anna::diameter::codec::functions::getApplicationId(db), ApplicationMessageOamModule::Counter::Request_Received_AsClient);

      eventRequest(db);
    } catch(anna::RuntimeException& ex) {
      ex.trace();
    }

    return;
  }

  /////////////////////////////
  // Here received an answer //
  /////////////////////////////
  bool doUnbind = false;
  bool immediateUnbind = false;
  int resultCode = -1;

  try {
    resultCode = helpers::base::functions::getResultCode(db);
  } catch(anna::RuntimeException& ex) {
    // Not always mandatory ...
    //ex.trace();
  }

  // Received CEA
  if(cid == helpers::base::COMMANDID__Capabilities_Exchange_Answer) {
    oamModule.count(OamModule::Counter::CEAReceived);

    if(a_state != State::WaitingBind) {
      LOGWARNING(anna::Logger::warning("Received CEA: unexpected message at not-WaitingBind state", ANNA_FILE_LOCATION));
      return; // we don't send its request
//         string msg("diameter::comm::ClientSession::receive | ");
//         msg += asString();
//         msg += " | Received CEA on not-WaitingBind state: unexpected Bind-response";
//         throw anna::RuntimeException(msg, ANNA_FILE_LOCATION);
    }

    if(resultCode != helpers::base::AVPVALUES__Result_Code::DIAMETER_SUCCESS) {
      LOGWARNING(
        std::string msg = anna::functions::asString("Received CEA with non-success Result-Code (%d). Unbinding connection.", resultCode);
        anna::Logger::warning(msg, ANNA_FILE_LOCATION);
      );
      doUnbind = true;
    } else {
      setState(State::Bound);
      //activateTimer(); // Ya se invoca al inicio de este metodo ::receive
      // Inform father server (availability changes):
      a_parent->refreshAvailability();
      //startClock();
    }
  }
  // Received DWA
  else if(cid == helpers::base::COMMANDID__Device_Watchdog_Answer) {
    oamModule.count(OamModule::Counter::DWAReceived);
    setWatchdogState(WatchdogState::WaitingTimerExpiration);

    if(resultCode != helpers::base::AVPVALUES__Result_Code::DIAMETER_SUCCESS)
      LOGWARNING(
        std::string msg = anna::functions::asString("Received DWA with non-success Result-Code (%d)... but ASSUME keep-alive is reached", resultCode);
        anna::Logger::warning(msg, ANNA_FILE_LOCATION);
      );

    if(a_state == State::Failover) {
      setState(State::Bound);
      LOGDEBUG(
        string msg("diameter::comm::ClientSession::receive | ");
        msg += asString();
        msg += " | Received DWA on failover state: recovering Bound state";
        anna::Logger::debug(msg, ANNA_FILE_LOCATION);
      );
    }

    // Keep-Alive don't manage context
    return;
  }
  // Received DPA
  else if(cid == helpers::base::COMMANDID__Disconnect_Peer_Answer) {
    oamModule.count(OamModule::Counter::DPAReceived);

    if(a_state == State::WaitingDPA) {
      if(resultCode != helpers::base::AVPVALUES__Result_Code::DIAMETER_SUCCESS) {
        LOGWARNING(
          std::string msg = anna::functions::asString("Received DPA with non-success Result-Code (%d). Ignoring and recovering Bound state", resultCode);
          anna::Logger::warning(msg, ANNA_FILE_LOCATION);
        );
        setState(State::Bound);
      } else {
        LOGWARNING(anna::Logger::warning("Received DPA With Result-Code = DIAMETER_SUCCESS. Disconnect now.", ANNA_FILE_LOCATION));
        immediateUnbind = true;
        doUnbind = true;
      }
    }
  }

  HopByHop hopByHop = codec::functions::getHopByHop(db); // context identification
  Response* response = response_find(hopByHop);

  // Out-of-context responses:
  if(!response) {
    // OAM
    oamModule.count(OamModule::Counter::AnswerReceivedUnknown);
    oamModule.count(OamModule::Counter::AnswerReceivedOnClientSessionUnknown);
    oamModule.activateAlarm(OamModule::Alarm::AnswerReceivedOnClientSessionUnknown);

    // application message counters
    ApplicationMessageOamModule::instantiate().count(cid.first, resultCode, anna::diameter::codec::functions::getApplicationId(db), ApplicationMessageOamModule::Counter::Answer_UnknownReceived_AsClient);

    eventUnknownResponse(db);
    string msg(asString());
    msg += anna::functions::asString(" | Response received from entity, for non registered context (HopByHop: %u)", hopByHop);
    throw anna::RuntimeException(msg, ANNA_FILE_LOCATION);
  }

  response->setResultCode(Response::ResultCode::Success);
  response->cancelTimer();
  LOGDEBUG(
    string msg("diameter::comm::ClientSession::receive | ");
    msg += asString();
    msg += " | Received answer";
    msg += response->asString();
    anna::Logger::debug(msg, ANNA_FILE_LOCATION);
  );
  // Statistics
  anna::Millisecond current = (anna::Millisecond)anna::functions::millisecond();
  anna::Millisecond request = response->getRequest()->getRequestTimestampMs();
  anna::Millisecond timeToAnswerMs = current - request;
  a_parent->updateProcessingTimeStatisticConcept(timeToAnswerMs, cid);
  //LOGDEBUG
  //(
  //  std::string msg = "This diameter request context lasted ";
  //  msg += anna::functions::asString(timeToAnswerMs);
  //  msg += " milliseconds at diameter server (included network time)";
  //  anna::Logger::debug(msg, ANNA_FILE_LOCATION);
  //);
  // Progress origin for tracking purposes on asyncronous boxes with both diameter interfaces (entities and clients)
  Message * requestMessage = const_cast<Message*>(response->getRequest());
  requestMessage->setRequestServerSessionKey(response->getRequest()->getRequestServerSessionKey()); // -1 means unkown/unset

  if(cid != helpers::base::COMMANDID__Disconnect_Peer_Answer) {
    try {
      response->setMessage(&db);
      // Restore received datablock
      LOGDEBUG(
        string msg("diameter::comm::ClientSession::receive | Restore answer to original request sequences (hop-by-hop = ");
        msg += anna::functions::asString(response->getRequest()->getRequestHopByHop());
        msg += ", end-to-end = ";
        msg += anna::functions::asString(response->getRequest()->getRequestEndToEnd());
        msg += ")";
        anna::Logger::debug(msg, ANNA_FILE_LOCATION);
      );
      diameter::codec::functions::setHopByHop((anna::DataBlock&)db, response->getRequest()->getRequestHopByHop());
      diameter::codec::functions::setEndToEnd((anna::DataBlock&)db, response->getRequest()->getRequestEndToEnd());

      // application message counters
      if(cid != helpers::base::COMMANDID__Capabilities_Exchange_Answer)
        ApplicationMessageOamModule::instantiate().count(cid.first, resultCode, anna::diameter::codec::functions::getApplicationId(db), ApplicationMessageOamModule::Counter::Answer_Received_AsClient);

      eventResponse(*response);
    } catch(anna::RuntimeException& ex) {
      ex.trace();
    }
  }
  else { // DPA
    // unbind is automatically performed, anyway we can inform to the application just in case some additional
    //  procedure could be issued:
    eventDPA(db);
  }

  response_erase(response);

  // Unbind trigger
  if(doUnbind)
    unbind(immediateUnbind);
}

void ClientSession::finalize() throw() {
  LOGMETHOD(anna::TraceMethod traceMethod(a_className, "finalize", ANNA_FILE_LOCATION));
  Session::finalize();
  // Inform father server (availability changes):
  a_parent->refreshAvailability();
  // OAM
  const Server *server = getParent();
  bool multipleConnections = (server->getMaxClientSessions() > 1);
  std::string socket = anna::functions::socketLiteralAsString(getAddress(), getPort());
  OamModule &oamModule = OamModule::instantiate();

  if(multipleConnections) {
    oamModule.activateAlarm(OamModule::Alarm::c_LostAvailabilityOverClientSessionWithServer__s__ClientSessionId__d__, socket.c_str(), getSocketId());
    oamModule.count(OamModule::Counter::LostAvailabilityOverClientSession);
  } else {
    oamModule.activateAlarm(OamModule::Alarm::c_LostAvailabilityOverClientSessionWithServer__s__, socket.c_str());
    oamModule.count(OamModule::Counter::LostAvailabilityOverClientSession);
  }
}

void ClientSession::recover() throw() {
  LOGMETHOD(anna::TraceMethod traceMethod(a_className, "recover", ANNA_FILE_LOCATION));

  try {
    bind();
  } catch(anna::RuntimeException &ex) {
    // Again:
    anna::comm::ClientSocket * cs = const_cast<anna::comm::ClientSocket*>(a_server->getClientSocket());

    if(cs) cs->requestClose();

    ex.trace();
  }

  // Inform father server (availability changes):
  a_parent->refreshAvailability();
  // OAM
  const Server *server = getParent();
  bool multipleConnections = (server->getMaxClientSessions() > 1);
  std::string socket = anna::functions::socketLiteralAsString(getAddress(), getPort());
  OamModule &oamModule = OamModule::instantiate();

  if(multipleConnections) {
    oamModule.cancelAlarm(OamModule::Alarm::c_LostAvailabilityOverClientSessionWithServer__s__ClientSessionId__d__, socket.c_str(), getSocketId());
    oamModule.count(OamModule::Counter::RecoveredAvailabilityOverClientSession);
  } else {
    oamModule.cancelAlarm(OamModule::Alarm::c_LostAvailabilityOverClientSessionWithServer__s__, socket.c_str());
    oamModule.count(OamModule::Counter::RecoveredAvailabilityOverClientSession);
  }
}

void ClientSession::sendDWAToServer(const anna::DataBlock& dwrDB)
throw(anna::RuntimeException) {
  LOGMETHOD(anna::TraceMethod traceMethod(a_className, "sendDWAToServer", ANNA_FILE_LOCATION));
  anna::DataBlock dwa(true);
  a_engine->readDWA(dwa, dwrDB); // Asume that DWA is valid ...

  if(dwa.isEmpty())
    throw anna::RuntimeException("This diameter agent defines an empty DWA message. Remote server never will validate this connection health", ANNA_FILE_LOCATION);

  Message msgDwa;
  msgDwa.setBody(dwa);
  send(&msgDwa);
}

//-------------------------------------------------------------------------
// Se invoca desde diameter::comm::Timer
//-------------------------------------------------------------------------
void ClientSession::expireResponse(diameter::comm::Response* response)
throw() {
  LOGMETHOD(anna::TraceMethod traceMethod(a_className, "expireResponse", ANNA_FILE_LOCATION));
  Session::expireResponse(response);
  // OAM
  OamModule &oamModule = OamModule::instantiate();
  oamModule.count(OamModule::Counter::RequestSentExpired);
  oamModule.count(OamModule::Counter::RequestSentOnClientSessionExpired);
  oamModule.activateAlarm(OamModule::Alarm::RequestSentOnClientSessionExpired);

  // Check father server idleness:
  if(idle()) a_parent->childIdle();

//   if (idle()) {
//      LOGDEBUG(anna::Logger::debug("ClientSession is idle after an expiration...", ANNA_FILE_LOCATION));
//      a_parent->childIdle();
//   }
//   else {
//      LOGDEBUG(anna::Logger::debug("ClientSession is busy after an expiration...", ANNA_FILE_LOCATION));
//   }
}


std::string ClientSession::asString() const
throw() {
  string result = Session::asString();
  result += " | Parent Server: ";
  result += anna::functions::socketLiteralAsString(getAddress(), getPort());
  result += " | Auto-recovery: ";
  result += (a_autoRecovery ? "yes" : "no");
  result += " | WatchdogState: ";
  result += asText(a_watchdogState);
  // Diferente del timeout de ApplicationMessage:
  result += " | Watchdog Period: ";
  result += getTimeout().asString();
  result += " | Hidden: ";
  result += (hidden() ? "yes" : "no");

  if(a_server) {
    result += " | MaxConnectionDelay: ";
    result += a_server->getMaxConnectionDelay().asString();
  }

  return result += " }";
}

anna::xml::Node* ClientSession::asXML(anna::xml::Node* parent) const
throw() {
  anna::xml::Node* result = Session::asXML(parent);
  parent->createChild("diameter.comm.ClientSession");
  result->createAttribute("ParentServer", anna::functions::socketLiteralAsString(getAddress(), getPort()));
  result->createAttribute("AutoRecovery", (a_autoRecovery ? "yes" : "no"));
  result->createAttribute("WatchdogState", asText(a_watchdogState));
  // Diferente del timeout de ApplicationMessage:
  result->createAttribute("WatchdogPeriod", getTimeout().asString());

  if(a_server) result->createAttribute("MaxConnectionDelay", a_server->getMaxConnectionDelay().asString());

  result->createAttribute("Hidden", hidden() ? "yes" : "no");
  return result;
}


const char* ClientSession::asText(const WatchdogState::_v watchdogState)
throw() {
  static const char* text [] = { "TimerStopped", "WaitingTimerExpiration", "WaitingDWA" };
  return text [watchdogState];
}


//------------------------------------------------------------------------------
//------------------------------------------------------ ClientSession::expire()
//------------------------------------------------------------------------------
void ClientSession::expire(anna::timex::Engine *timeController) throw(anna::RuntimeException) {
  LOGMETHOD(anna::TraceMethod traceMethod(a_className, "expire (watchdog timer)", ANNA_FILE_LOCATION));

  //   The client MUST NOT close the primary connection until the
  //    primary's watchdog timer has expired at least twice without a
  //    response (note that the watchdog is not sent a second time,
  //    however).
  if(a_watchdogState == WatchdogState::WaitingDWA) {
    if(a_state == State::Failover) {
      LOGWARNING(anna::Logger::warning("Unbinding client-session: Tw expired after first DWA missing (2*Tw elapsed)", ANNA_FILE_LOCATION));
      unbind();
      return; // finalize will stop the stopped timer ...
    }

    setState(State::Failover);
    LOGWARNING(anna::Logger::warning("Going to Failover state: first DWA missing", ANNA_FILE_LOCATION));
    activateTimer(); // another chance on failover
    return;
  }

  // WaitingTimerExpiration arrives here:
  const Response* sent;

  try {
    sent = send(&a_dwr);
  } catch(anna::RuntimeException&) {
    LOGDEBUG(anna::Logger::debug("Failed to send DWR to the server: unbinding ...", ANNA_FILE_LOCATION));
    setState(State::Failover);
    unbind();
    throw;
  }

  LOGDEBUG(if(sent) anna::Logger::debug("DWR sent to the server", ANNA_FILE_LOCATION););

  activateTimer();
}

void ClientSession::setWatchdogPeriod(const anna::Millisecond & watchdogPeriod) throw() {
  setTimeout(watchdogPeriod);
}

void ClientSession::setWatchdogState(WatchdogState::_v wState) throw() {
  LOGDEBUG(

  if(wState != a_watchdogState) {
  std::string msg("Session watchdog state change: ");
    msg += asText(a_watchdogState);
    msg += " -> ";
    msg += asText(wState);
    anna::Logger::debug(msg, ANNA_FILE_LOCATION);
  }
  );
  a_watchdogState = wState;
}


void ClientSession::timerStopped() throw() {
  LOGDEBUG(anna::Logger::debug("Watchdog timer stopped", ANNA_FILE_LOCATION));
  setWatchdogState(WatchdogState::TimerStopped);
}

void ClientSession::timerStarted() throw() {
  LOGDEBUG(anna::Logger::debug("Watchdog timer started", ANNA_FILE_LOCATION));

  if(a_watchdogState == WatchdogState::WaitingDWA) return;

  setWatchdogState(WatchdogState::WaitingTimerExpiration);
}


//------------------------------------------------------------------------------
//---------------------------------- ClientSession::updateIncomingActivityTime()
//------------------------------------------------------------------------------
void ClientSession::updateIncomingActivityTime() throw() {
  Session::updateIncomingActivityTime();
  a_parent->updateIncomingActivityTime();
}


//------------------------------------------------------------------------------
//---------------------------------- ClientSession::updateOutgoingActivityTime()
//------------------------------------------------------------------------------
void ClientSession::updateOutgoingActivityTime(void) throw() {
  Session::updateOutgoingActivityTime();
  a_parent->updateOutgoingActivityTime();
}



//------------------------------------------------------------------------------
//----------------------------------------------- ClientSession::countSendings()
//------------------------------------------------------------------------------
void ClientSession::countSendings(const diameter::CommandId & cid, unsigned int aid, bool ok)throw() {
  OamModule &oamModule = OamModule::instantiate();
  ApplicationMessageOamModule &appMsgOamModule = ApplicationMessageOamModule::instantiate();

  bool isRequest = cid.second;

  if(ok) {
    // Main counters:
    oamModule.count(isRequest ? OamModule::Counter::RequestSentOK : OamModule::Counter::AnswerSentOK);
    oamModule.count(isRequest ? OamModule::Counter::RequestSentOnClientSessionOK : OamModule::Counter::AnswerSentOnClientSessionOK);

    if(cid == helpers::base::COMMANDID__Capabilities_Exchange_Request) oamModule.count(OamModule::Counter::CERSentOK);
    else if(cid == helpers::base::COMMANDID__Device_Watchdog_Request) oamModule.count(OamModule::Counter::DWRSentOK);
    else if(cid == helpers::base::COMMANDID__Device_Watchdog_Answer) oamModule.count(OamModule::Counter::DWASentOK);  // not usual (dwr was received from server)
    else if(cid == helpers::base::COMMANDID__Disconnect_Peer_Request) oamModule.count(OamModule::Counter::DPRSentOK);
    else if(cid == helpers::base::COMMANDID__Disconnect_Peer_Answer) oamModule.count(OamModule::Counter::DPASentOK);
    // Application messages:
    else {
      appMsgOamModule.count(cid.first, -1 /* no result code */, aid, isRequest ? ApplicationMessageOamModule::Counter::Request_SentOK_AsClient : ApplicationMessageOamModule::Counter::Answer_SentOK_AsClient);
    }
  } else {
    // Main counters:
    oamModule.count(isRequest ? OamModule::Counter::RequestSentNOK : OamModule::Counter::AnswerSentNOK);
    oamModule.count(isRequest ? OamModule::Counter::RequestSentOnClientSessionNOK : OamModule::Counter::AnswerSentOnClientSessionNOK);

    if(cid == helpers::base::COMMANDID__Capabilities_Exchange_Request) oamModule.count(OamModule::Counter::CERSentNOK);
    else if(cid == helpers::base::COMMANDID__Device_Watchdog_Request) oamModule.count(OamModule::Counter::DWRSentNOK);
    else if(cid == helpers::base::COMMANDID__Device_Watchdog_Answer) oamModule.count(OamModule::Counter::DWASentNOK);  // not usual (dwr was received from server)
    else if(cid == helpers::base::COMMANDID__Disconnect_Peer_Request) oamModule.count(OamModule::Counter::DPRSentNOK);
    else if(cid == helpers::base::COMMANDID__Disconnect_Peer_Answer) oamModule.count(OamModule::Counter::DPASentNOK);
    // Application messages:
    else {
      appMsgOamModule.count(cid.first, -1 /* no result code */, aid, isRequest ? ApplicationMessageOamModule::Counter::Request_SentNOK_AsClient : ApplicationMessageOamModule::Counter::Answer_SentNOK_AsClient);
    }
  }
}

